/*
 * Copyright (c) 2017 Adrian Siekierka
 *
 * This file is part of Unlimited Chisel Works.
 *
 * Unlimited Chisel Works is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Unlimited Chisel Works is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Unlimited Chisel Works.  If not, see <http://www.gnu.org/licenses/>.
 */

package pl.asie.ucw;

import com.google.common.base.Charsets;
import com.google.common.collect.ImmutableSet;
import com.google.gson.*;
import net.minecraft.client.Minecraft;
import net.minecraft.client.resources.IResource;
import net.minecraft.client.resources.IResourceManager;
import net.minecraft.client.resources.IResourceManagerReloadListener;
import net.minecraft.client.resources.IResourcePack;
import net.minecraft.client.resources.data.IMetadataSection;
import net.minecraft.client.resources.data.MetadataSerializer;
import net.minecraft.util.JsonUtils;
import net.minecraft.util.ResourceLocation;

import javax.annotation.Nullable;
import java.awt.image.BufferedImage;
import java.io.*;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.function.Consumer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class UCWFakeResourcePack implements IResourcePack, IResourceManagerReloadListener {
	public static final UCWFakeResourcePack INSTANCE = new UCWFakeResourcePack();

	private final Map<ResourceLocation, byte[]> data = new HashMap<>();
	private final Map<ResourceLocation, JsonElement> jsonCache = new HashMap<>();
	private final Set<String> domains = ImmutableSet.of("ucw_generated");

	private UCWFakeResourcePack() {

	}

	public JsonElement parseJsonElement(String[] str, JsonElement element) {
		if (element.isJsonObject()) {
			JsonObject parent = element.getAsJsonObject();
			JsonObject object = new JsonObject();
			for (Map.Entry<String, JsonElement> entry : parent.entrySet()) {
				object.add(entry.getKey(), parseJsonElement(str, entry.getValue()));
			}
			return object;
		} else if (element.isJsonArray()) {
			JsonArray parent = element.getAsJsonArray();
			JsonArray array = new JsonArray();
			for (JsonElement element1 : parent) {
				array.add(parseJsonElement(str, element1));
			}
			return array;
		} else if (!element.isJsonNull()) {
			String s = element.getAsString();
			if (s != null && s.startsWith(str[1] + ":")) {
				String ns = "ucw_generated:ucw_ucw_" + str[0] + "/" + s.replaceFirst(":", "/");
				// System.out.println(s + " -> " + ns);
				return new JsonPrimitive(ns);
			} else {
				return element;
			}
		} else {
			return element;
		}
	}

	private String[] getStr(ResourceLocation location) {
		String path = location.func_110623_a();
		if (path.contains(".png")) {
			return null;
		}
		Matcher m = Pattern.compile("ucw_ucw_([A-Za-z0-9_]+)/([a-z]+)").matcher(path);
		String[] str = new String[] {
				"",
				"",
				""
		};

		if (m.find()) {
			str[0] = m.group(1);
			str[1] = m.group(2);
		}

		String nonProxyPath = path.replaceAll("ucw_ucw_[A-Za-z0-9_]+/[a-z]+/", "");
		str[2] = nonProxyPath;
		return str;
	}

	@Override
	public InputStream func_110590_a(ResourceLocation location) throws IOException {
		if (!data.containsKey(location)) {
			ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();

			String[] str = getStr(location);
			JsonElement element;
			ResourceLocation nonProxiedLoc = new ResourceLocation(str[1], str[2]);

			if (jsonCache.containsKey(nonProxiedLoc)) {
				element = jsonCache.get(nonProxiedLoc);
			} else {
				InputStream nonProxied = Minecraft.func_71410_x().func_110442_L().func_110536_a(
						nonProxiedLoc
				).func_110527_b();
				Reader reader = new InputStreamReader(nonProxied);

				try {
					element = UnlimitedChiselWorks.GSON.fromJson(reader, JsonElement.class);
				} catch (Exception e) {
					element = null;
					reader.close();
					nonProxied.close();
				}
			}

			if (element != null) {
				JsonElement newElement = parseJsonElement(str, element);
				byte[] b = UnlimitedChiselWorks.GSON.toJson(newElement).getBytes(Charsets.UTF_8);
				byteArrayOutputStream.write(b);
			} else {
				InputStream nonProxied = Minecraft.func_71410_x().func_110442_L().func_110536_a(
						new ResourceLocation(str[1], str[2])
				).func_110527_b();

				while (nonProxied.available() > 0)
					byteArrayOutputStream.write(nonProxied.read());

				nonProxied.close();
			}

			byteArrayOutputStream.close();

			byte[] out = byteArrayOutputStream.toByteArray();
			data.put(location, out);
			return new ByteArrayInputStream(out);
		} else {
			return new ByteArrayInputStream(data.get(location));
		}
	}

	@Override
	public boolean func_110589_b(ResourceLocation location) {
		String[] str = getStr(location);
		if (str == null || str[1].isEmpty()) return false;

		try {
			IResource resource = Minecraft.func_71410_x().func_110442_L().func_110536_a(
					new ResourceLocation(str[1], str[2])
			);
			return true;
		} catch (Exception e) {
			return false;
		}
	}

	@Override
	public Set<String> func_110587_b() {
		return domains;
	}

	@Nullable
	@Override
	public <T extends IMetadataSection> T func_135058_a(MetadataSerializer metadataSerializer, String metadataSectionName) throws IOException {
		return null;
	}

	@Override
	public BufferedImage func_110586_a() throws IOException {
		return null;
	}

	@Override
	public String func_130077_b() {
		return "UCWFakePack";
	}

	@Override
	public void func_110549_a(IResourceManager resourceManager) {
		data.clear();
		jsonCache.clear();
	}
}
